--                                                                    --
--  package APQ.Common              Copyright (c)  Dmitry A. Kazakov  --
--  Implementation                                 Luebeck            --
--                                                 Winter, 2004       --
--                                                                    --
--                                Last revision :  12:32 10 Jun 2003  --
--                                                                    --
--  This  library  is  free software; you can redistribute it and/or  --
--  modify it under the terms of the GNU General Public  License  as  --
--  published by the Free Software Foundation; either version  2  of  --
--  the License, or (at your option) any later version. This library  --
--  is distributed in the hope that it will be useful,  but  WITHOUT  --
--  ANY   WARRANTY;   without   even   the   implied   warranty   of  --
--  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU  --
--  General  Public  License  for  more  details.  You  should  have  --
--  received  a  copy  of  the GNU General Public License along with  --
--  this library; if not, write to  the  Free  Software  Foundation,  --
--  Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.    --
--                                                                    --
--  As a special exception, if other files instantiate generics from  --
--  this unit, or you link this unit with other files to produce  an  --
--  executable, this unit does not by  itself  cause  the  resulting  --
--  executable to be covered by the GNU General Public License. This  --
--  exception  does not however invalidate any other reasons why the  --
--  executable file might be covered by the GNU Public License.       --
--____________________________________________________________________--

with Ada.Characters.Handling;  use Ada.Characters.Handling;
with Ada.Exceptions;           use Ada.Exceptions;
with Ada.IO_Exceptions;        use Ada.IO_Exceptions;
with Strings_Edit;             use Strings_Edit;
with Strings_Edit.Integers;    use Strings_Edit.Integers;

with APQ.MySQL.Client;
with APQ.PostgreSQL.Client;

package body APQ.Common is

   procedure Free is new
      Ada.Unchecked_Deallocation
      (  Root_Connection_Type'Class,
         Root_Connection_Ptr
      );
   procedure Free is new
      Ada.Unchecked_Deallocation
      (  Root_Query_Type'Class,
         Root_Query_Ptr
      );

   procedure Append_Quoted
             (  Data_Base : in out APQ_Data_Base;
                Text      : String;
                After     : String := ""
             )  is
   begin
      Append_Quoted
      (  Data_Base.Query.all,
         Data_Base.Connection.all,
         Text,
         After
      );
   end Append_Quoted;

   function Encode (Text : Wide_String; Prefix : String)
      return String is
      Result  : String  := Prefix & (1.. 2 * Text'Length => ' ');
      Pointer : Integer := Result'First + Prefix'Length;
      Index   : Integer := Text'First;
   begin
      while Index <= Text'Last loop
         case Text (Index) is
            when ' '..'!' | '$'..'%' | '('..';' | '=' | '?'..'~' =>
               Put (Result, Pointer, To_Character (Text (Index)));
            when others =>
               declare
                  Position : Integer := Pointer;
               begin
                  Put (Result, Position, "&#");
                  Put
                  (  Result,
                     Position,
                     Wide_Character'Pos (Text (Index))
                  );
                  Put (Result, Position, ';');
                  Pointer := Position;
               end;
         end case;
         Index := Index + 1;
      end loop;
      return Result (Result'First..Pointer - 1);
   exception
      when Layout_Error =>
         return
            Encode
            (  Text (Index..Text'Last),
               Result (Result'First..Pointer - 1)
            );
   end Encode;

   procedure Append_Quoted
             (  Data_Base : in out APQ_Data_Base;
                Text      : Wide_String;
                After     : String := ""
             )  is
   begin
      Append_Quoted
      (  Data_Base.Query.all,
         Data_Base.Connection.all,
         Encode (Text, ""),
         After
      );
   end Append_Quoted;

   procedure Close_DB_Trace (Data_Base : in out APQ_Data_Base) is
   begin
      case Engine_Of (Data_Base.Connection.all) is
         when Engine_PostgreSQL =>
            declare
               use APQ.PostgreSQL.Client;
               Object : Connection_Type renames
                           Connection_Type (Data_Base.Connection.all);
            begin
               if Is_Trace (Object) then
                  APQ.PostgreSQL.Client.Close_DB_Trace (Object);
               end if;
            end;
         when Engine_MySQL =>
            declare
               use APQ.MySQL.Client;
               Object : Connection_Type renames
                           Connection_Type (Data_Base.Connection.all);
            begin
               if Is_Trace (Object) then
                  APQ.MySQL.Client.Close_DB_Trace (Object);
               end if;
            end;
         when others =>
            null;
      end case;      
   end Close_DB_Trace;

   procedure Commit (Data_Base : in out APQ_Data_Base) is
   begin
      case Data_Base.Shared is
         when Fully => 
            Raise_Exception
            (  Use_Error'Identity,
               "Multiple commits are not allowed"
            );
         when Read_Only =>
            Data_Base.Shared := Fully;
         when Read_Write =>
            Data_Base.Shared := Fully;
            Commit_Work (Data_Base.Query.all, Data_Base.Connection.all);
      end case;
   exception
      when Use_Error =>
         raise;
      when SQL_Error =>
         Raise_Exception
         (  Data_Error'Identity,
            Error_Message (Data_Base.Query.all)
         );
      when Error : others =>
         Raise_Exception
         (  Data_Error'Identity,
            Exception_Message (Error)
         );
   end Commit;

   procedure Connect
             (  Data_Base      : in out APQ_Data_Base;
                Server_Type    : Database_Type;
                Data_Base_Name : String;
                User_Name      : String;
                Password       : String;
                Host_Name      : String  := "localhost";
                Port_Number    : Natural := 0
             )  is
   begin
      case Server_Type is
         when Engine_PostgreSQL =>
            Data_Base.Connection :=
               new PostgreSQL.Client.Connection_Type;
         when Engine_MySQL =>
            Data_Base.Connection :=
               new MySQL.Client.Connection_Type;
         when others =>
            Raise_Exception
            (  Use_Error'Identity,
               (  Database_Type'Image (Server_Type)
               &  " data bases are not yet supported by APQ"
            )  );
      end case;
      Set_Host_Name (Data_Base.Connection.all, Host_Name);
      if Port_Number > 0 then
         Set_Port (Data_Base.Connection.all, Port_Number);
      end if;
      Set_DB_Name (Data_Base.Connection.all, Data_Base_Name);
      Set_User_Password (Data_Base.Connection.all, User_Name, Password);
         Connect (Data_Base.Connection.all);
      Data_Base.Query :=
         new Root_Query_Type'Class'
             (  New_Query (Data_Base.Connection.all)
             );
   exception
      when Not_Connected =>
         Raise_Exception
         (  Use_Error'Identity,
            Error_Message (Data_Base.Connection.all)
         );
   end Connect;

   procedure Drop
             (  Data_Base  : in out APQ_Data_Base;
                Table_Name : String
             )  is
   begin
      Seize_Write (Data_Base);
      Prepare (Data_Base.Query.all, "DROP TABLE " & Table_Name);
      Execute (Data_Base);
      Commit (Data_Base);
   exception
      when Data_Error =>
         Roll_Back (Data_Base);
      when others =>
         raise;
   end Drop;

   procedure Execute (Data_Base : APQ_Data_Base) is
   begin
      if Data_Base.Shared = Fully then
         Raise_Exception
         (  Use_Error'Identity,
            "No access to execute"
         );
      end if;
      Execute (Data_Base.Query.all, Data_Base.Connection.all);
   exception
      when Use_Error =>
         raise;
      when SQL_Error =>
         Raise_Exception
         (  Data_Error'Identity,
            Error_Message (Data_Base.Query.all)
         );
      when Error : others =>
         Raise_Exception
         (  Data_Error'Identity,
            Exception_Message (Error)
         );
   end Execute;

   procedure Finalize (Data_Base : in out APQ_Data_Base) is
   begin
      Free (Data_Base.Query);
      Free (Data_Base.Connection);
   end Finalize;

   function Get_ID
            (  Data_Base   : APQ_Data_Base;
               Table_Name  : String;
               Column_Name : String
            )  return Row_ID_Type is
   begin
      case Engine_Of (Data_Base.Connection.all) is
         when Engine_PostgreSQL =>
            declare
               ID : constant Row_ID_Type :=
                       Command_OID (Data_Base.Query.all);
            begin
               Prepare
               (  Data_Base.Query.all,
                  (  "SELECT "
                  &  Column_Name
                  &  " FROM "
                  &  Table_Name
                  & " WHERE oid="
               )  );
               Append (Data_Base.Query.all, ID);
               Execute
               (  Data_Base.Query.all,
                  Data_Base.Connection.all
               );
               Fetch (Data_Base.Query.all);
               return Value (Data_Base.Query.all, 1);
            end;
         when Engine_MySQL =>
            return Command_OID (Data_Base.Query.all);
         when others =>
            Raise_Exception
            (  Use_Error'Identity,
               (  Database_Type'Image
                  (  Engine_Of (Data_Base.Connection.all)
                  )
               &  " does not support row identification"
            )  );
      end case;
   exception
      when No_Tuple =>
         Raise_Exception (Data_Error'Identity, "Insertion failed");
      when Data_Error | Use_Error =>
         raise;
      when SQL_Error =>
         Raise_Exception
         (  Data_Error'Identity,
            Error_Message (Data_Base.Query.all)
         );
      when Error : others =>
         Raise_Exception
         (  Data_Error'Identity,
            Exception_Message (Error)
         );
   end Get_ID; 

   function ID_Type (Data_Base : APQ_Data_Base) return String is
   begin
      case Engine_Of (Data_Base.Connection.all) is
         when Engine_PostgreSQL =>
            return "SERIAL";
         when Engine_MySQL =>
            return "INTEGER NOT NULL AUTO_INCREMENT PRIMARY KEY";
         when others =>
            Raise_Exception
            (  Use_Error'Identity,
               (  Database_Type'Image
                  (  Engine_Of (Data_Base.Connection.all)
                  )
               &  " does not support row identification"
            )  );
      end case;
   end ID_Type; 

   function Integer_Type (Data_Base : APQ_Data_Base) return String is
   begin
      case Engine_Of (Data_Base.Connection.all) is
         when Engine_PostgreSQL =>
            return "INTEGER";
         when Engine_MySQL =>
            return "INTEGER";
         when others =>
            Raise_Exception
            (  Use_Error'Identity,
               (  Database_Type'Image
                  (  Engine_Of (Data_Base.Connection.all)
                  )
               &  " does not support integers"
            )  );
      end case;
   end Integer_Type; 

   function New_ID (Data_Base : APQ_Data_Base) return String is
   begin
      case Engine_Of (Data_Base.Connection.all) is
         when Engine_PostgreSQL =>
            return "DEFAULT";
         when Engine_MySQL =>
            return "NULL";
         when others =>
            Raise_Exception
            (  Use_Error'Identity,
               (  Database_Type'Image
                  (  Engine_Of (Data_Base.Connection.all)
                  )
               &  " does not support row identification"
            )  );
      end case;
   end New_ID; 

   procedure Open_DB_Trace
             (  Data_Base : APQ_Data_Base;
                File_Name : String;
                Mode      : Trace_Mode_Type := Trace_APQ
             )  is
   begin
      case Engine_Of (Data_Base.Connection.all) is
         when Engine_PostgreSQL =>
            APQ.PostgreSQL.Client.Open_DB_Trace
            (  APQ.PostgreSQL.Client.Connection_Type
               (  Data_Base.Connection.all
               ),
               File_Name,
               Mode
            );
         when Engine_MySQL =>
            APQ.MySQL.Client.Open_DB_Trace
            (  APQ.MySQL.Client.Connection_Type
               (  Data_Base.Connection.all
               ),
               File_Name,
               Mode
            );
         when others =>
            Raise_Exception
            (  Use_Error'Identity,
               (  Database_Type'Image
                  (  Engine_Of (Data_Base.Connection.all)
                  )
               &  " does not support tracing"
            )  );
      end case;
   end Open_DB_Trace;

   function Ref_Type (Data_Base : APQ_Data_Base) return String is
   begin
      case Engine_Of (Data_Base.Connection.all) is
         when Engine_PostgreSQL =>
            return "INTEGER";
         when Engine_MySQL =>
            return "INTEGER";
         when others =>
            Raise_Exception
            (  Use_Error'Identity,
               (  Database_Type'Image
                  (  Engine_Of (Data_Base.Connection.all)
                  )
               &  " does not support row identification"
            )  );
      end case;
   end Ref_Type; 

   procedure Roll_Back (Data_Base : in out APQ_Data_Base) is
   begin
      case Data_Base.Shared is
         when Fully => 
            null;
         when Read_Only =>
            Data_Base.Shared := Fully;
         when Read_Write =>
            Data_Base.Shared := Fully;
            RollBack_Work
            (  Data_Base.Query.all,
               Data_Base.Connection.all
            );
      end case;
   exception
      when Use_Error =>
         raise;
      when SQL_Error =>
         Raise_Exception
         (  Data_Error'Identity,
            Error_Message (Data_Base.Query.all)
         );
      when Error : others =>
         Raise_Exception
         (  Data_Error'Identity,
            Exception_Message (Error)
         );
   end Roll_Back;

   procedure Seize_Read (Data_Base : in out APQ_Data_Base) is
   begin
      case Data_Base.Shared is
         when Fully => 
            Data_Base.Shared := Read_Only;
         when Read_Only | Read_Write =>
            Raise_Exception
            (  Use_Error'Identity,
               "Nested read transaction"
            );
      end case;
   end Seize_Read;

   procedure Seize_Write (Data_Base : in out APQ_Data_Base) is
   begin
      case Data_Base.Shared is
         when Fully =>
            Data_Base.Shared := Read_Write;
            Begin_Work
            (  Data_Base.Query.all,
               Data_Base.Connection.all
            );
         when Read_Only | Read_Write =>
            Raise_Exception
            (  Use_Error'Identity,
               "Nested write transaction"
            );
      end case;
   end Seize_Write;

   function String_Type (Data_Base : APQ_Data_Base) return String is
   begin
      case Engine_Of (Data_Base.Connection.all) is
         when Engine_PostgreSQL =>
            return "TEXT";
         when Engine_MySQL =>
            return "MEDIUMTEXT";
         when others =>
            Raise_Exception
            (  Use_Error'Identity,
               (  Database_Type'Image
                  (  Engine_Of (Data_Base.Connection.all)
                  )
               &  " does not support varying strings"
            )  );
      end case;
   end String_Type; 

   function Table_Exists
            (  Data_Base  : APQ_Data_Base;
               Table_Name : String
            )  return Boolean is
   begin
      case Data_Base.Shared is
         when Fully | Read_Write =>
            Raise_Exception
            (  Use_Error'Identity,
               "Read (but write) transaction required for Table_Exists"
            );
         when Read_Only =>
            null;
      end case;
      Prepare (Data_Base.Query.all, "SELECT * FROM " & Table_Name);
      Execute (Data_Base);
      Clear (Data_Base.Query.all);
      return True;
   exception
      when Use_Error =>
         raise;
      when others =>
         return False;
   end Table_Exists;

   function Timestamp_Now (Data_Base : APQ_Data_Base) return String is
   begin
      case Engine_Of (Data_Base.Connection.all) is
         when Engine_PostgreSQL =>
            return "now()";
         when Engine_MySQL =>
            return "now()";
         when others =>
            Raise_Exception
            (  Use_Error'Identity,
               (  Database_Type'Image
                  (  Engine_Of (Data_Base.Connection.all)
                  )
               &  " does not support timestamps"
            )  );
      end case;
   end Timestamp_Now; 

   function Timestamp_Type (Data_Base : APQ_Data_Base) return String is
   begin
      case Engine_Of (Data_Base.Connection.all) is
         when Engine_PostgreSQL =>
            return "TIMESTAMP";
         when Engine_MySQL =>
            return "TIMESTAMP";
         when others =>
            Raise_Exception
            (  Use_Error'Identity,
               (  Database_Type'Image
                  (  Engine_Of (Data_Base.Connection.all)
                  )
               &  " does not support timestamps"
            )  );
      end case;
   end Timestamp_Type; 

   function Value
            (  Data_Base : APQ_Data_Base;
               Column    : Column_Index_Type
            )  return Wide_String is
      Source      : String renames Value (Data_Base.Query.all, Column);
      Destination : Wide_String (1..Source'Length);
      From        : Integer := Source'First;
      To          : Integer := Destination'First;
      Code        : Integer;
   begin
      while From <= Source'Last loop
         if '&' = Source (From) then
            From := From + 1;
            if From > Source'Last or else '#' /= Source (From) then
               raise Data_Error;
            end if;
            From := From + 1;
            Get (Source, From, Code);
            if From > Source'Last or else ';' /= Source (From) then
               raise Data_Error;
            end if;
            Destination (To) := Wide_Character'Val (Code);
         else
            Destination (To) := To_Wide_Character (Source (From));
         end if;
         From := From + 1;
         To   := To   + 1;
      end loop;                  
      return Destination (1..To - 1);
   exception
      when Data_Error | End_Error | Layout_Error =>
         Raise_Exception
         (  Data_Error'Identity,
            "Illegal &#<number>; sequence"
         );
      when Error : others =>
         Raise_Exception
         (  Data_Error'Identity,
            Exception_Message (Error)
         );
   end Value;

   function Value
            (  Data_Base : APQ_Data_Base;
               Column    : Column_Index_Type
            )  return String is
   begin
      return Value (Data_Base.Query.all, Column);
   exception
      when Error : others =>
         Raise_Exception
         (  Data_Error'Identity,
            Exception_Message (Error)
         );
   end Value;

end APQ.Common;
