--                                                                    --
--  package                         Copyright (c)  Dmitry A. Kazakov  --
--     Object.Handle.Generic_Bounded_Array         Luebeck            --
--  Interface                                      Spring, 2003       --
--                                                                    --
--                                Last revision :  22:07 05 Dec 2004  --
--                                                                    --
--  This  library  is  free software; you can redistribute it and/or  --
--  modify it under the terms of the GNU General Public  License  as  --
--  published by the Free Software Foundation; either version  2  of  --
--  the License, or (at your option) any later version. This library  --
--  is distributed in the hope that it will be useful,  but  WITHOUT  --
--  ANY   WARRANTY;   without   even   the   implied   warranty   of  --
--  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU  --
--  General  Public  License  for  more  details.  You  should  have  --
--  received  a  copy  of  the GNU General Public License along with  --
--  this library; if not, write to  the  Free  Software  Foundation,  --
--  Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.    --
--                                                                    --
--  As a special exception, if other files instantiate generics from  --
--  this unit, or you link this unit with other files to produce  an  --
--  executable, this unit does not by  itself  cause  the  resulting  --
--  executable to be covered by the GNU General Public License. This  --
--  exception  does not however invalidate any other reasons why the  --
--  executable file might be covered by the GNU Public License.       --
--____________________________________________________________________--
--
--  This  package  defines  a generic type Bounded_Array. An instance of
--  Bounded_Array is  a  bounded  array  of  pointers  to  automatically
--  collected  objects  (see  the  package  Object). The package has the
--  following generic parameters:
--
--     Index_Type  - The array index type
--     Handle_Type - A handle type derived from Handle
--
with Ada.Finalization;

generic
   type Index_Type is (<>);
   type Handle_Type is new Handle with private;
package Object.Handle.Generic_Bounded_Array is
   type Bounded_Array (First, Last : Index_Type) is
      new Ada.Finalization.Controlled with private;
--
-- Adjust -- Assignment (part of)
--
--    Container - The array
--
   procedure Adjust (Container : in out Bounded_Array);
   pragma Inline (Adjust);
--
-- Finalize -- Destructor
--
--    Container - The array
--
   procedure Finalize (Container : in out Bounded_Array);
   pragma Inline (Finalize);
--
-- Fill -- Replace an array slice
--
--    Container - The array
--    From      - Index of the element in the slice
--    To        - Index of the last element
--    Element   - The replacement (a pointer/handle to)
--
-- Nothing  happens if From > To. Otherwise both have to be in the array
-- index range.
--
-- Exceptions :
--
--    Constraint_Error - Illegal index
--
   procedure Fill
             (  Container : in out Bounded_Array;
                From      : Index_Type;
                To        : Index_Type;
                Element   : Object_Type_Ptr
             );
   procedure Fill
             (  Container : in out Bounded_Array;
                From      : Index_Type;
                To        : Index_Type;
                Element   : Handle_Type
             );
   pragma Inline (Fill);
--
-- Get -- Get an array element by its index
--
--    Container - The array
--    Index     - Of the element
--
-- Returns :
--
--    The element or null if none
--
   function Get
            (  Container : Bounded_Array;
               Index     : Index_Type
            )  return Object_Type_Ptr;
   pragma Inline (Get);
--
-- Put -- Replace an array element by its index
--
--    Container - The array
--    Index     - Of the element
--    Element   - The new element (a pointer/handle to)
--
-- Exceptions :
--
--    Constraint_Error - Illegal index
--
   procedure Put
             (  Container : in out Bounded_Array;
                Index     : Index_Type;
                Element   : Object_Type_Ptr
             );
   procedure Put
             (  Container : in out Bounded_Array;
                Index     : Index_Type;
                Element   : Handle_Type
             );
   pragma Inline (Put);
--
-- Ref -- Get a handle to an array element by its index
--
--    Container - The array
--    Index     - Of the element
--
-- Returns :
--
--    A handle to the element
--
-- Exceptions :
--
--    Constraint_Error - No such element
--
   function Ref
            (  Container : Bounded_Array;
               Index     : Index_Type
            )  return Handle_Type;
   pragma Inline (Ref);

private
   type Object_Ptr_Array_Type is
      array (Index_Type range <>) of Object_Type_Ptr;

   type Bounded_Array (First, Last : Index_Type) is
      new Ada.Finalization.Controlled with
   record
      Vector : Object_Ptr_Array_Type (First..Last);
   end record;

end Object.Handle.Generic_Bounded_Array;
