--                                                                    --
--  package                         Copyright (c)  Dmitry A. Kazakov  --
--     Persistent.Data_Bank.Index                  Luebeck            --
--  Implementation                                 Autumn, 2004       --
--                                                                    --
--                                Last revision :  12:32 10 Jun 2003  --
--                                                                    --
--  This  library  is  free software; you can redistribute it and/or  --
--  modify it under the terms of the GNU General Public  License  as  --
--  published by the Free Software Foundation; either version  2  of  --
--  the License, or (at your option) any later version. This library  --
--  is distributed in the hope that it will be useful,  but  WITHOUT  --
--  ANY   WARRANTY;   without   even   the   implied   warranty   of  --
--  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU  --
--  General  Public  License  for  more  details.  You  should  have  --
--  received  a  copy  of  the GNU General Public License along with  --
--  this library; if not, write to  the  Free  Software  Foundation,  --
--  Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.    --
--                                                                    --
--  As a special exception, if other files instantiate generics from  --
--  this unit, or you link this unit with other files to produce  an  --
--  executable, this unit does not by  itself  cause  the  resulting  --
--  executable to be covered by the GNU General Public License. This  --
--  exception  does not however invalidate any other reasons why the  --
--  executable file might be covered by the GNU Public License.       --
--____________________________________________________________________--

with Ada.IO_Exceptions;  use Ada.IO_Exceptions;

with Ada.Unchecked_Deallocation;

package body Persistent.Data_Bank.Index is
   use Ptr_Map;
   use Key_Map;
   use Name_Map;
--
-- Create -- New catalogue record
--
--    List     - The catalogue
--    Object   - The object
--    External - The object's key
--    Name     - The object's name
--
   procedure Create
             (  List      : in out Catalogue;
                Object    : Deposit_Ptr;
                External  : Key;
                Name      : Wide_String
             );
--
-- Delete -- A catalogue record
--
--    List - The catalogue
--    Item - The record to be deleted (a pointer to)
--   
   procedure Delete
             (  List : in out Catalogue;
                Item : Catalogue_Record_Ptr
             );

   procedure Free is
      new Ada.Unchecked_Deallocation
          (  Catalogue_Record,
             Catalogue_Record_Ptr
          );

   function "<" (Left, Right : By_Name_Ptr) return Boolean is
   begin
      if Equal (Right, null) then
         return False;
      elsif Equal (Left, null) then
         return True;
      else
         return Left.Name < Right.Name;
      end if;
   end "<";

   function "<" (Left, Right : By_Entity_Ptr) return Boolean is
   begin
      if Equal (Right, null) then
         return False;
      elsif Equal (Left, null) then
         return True;
      else
         if Left.all in Catalogue_Record'Class then
            if Right.all in Catalogue_Record'Class then
               return
                  Less
                  (  This (Catalogue_Record'Class (Left.all )).all,
                     This (Catalogue_Record'Class (Right.all)).all
                  );
            else
               return
                  Less
                  (  This (Catalogue_Record'Class (Left.all)).all,
                     Right.all
                  );
            end if;
         else
            if Right.all in Catalogue_Record'Class then
               return
                  Object.Less
                  (  Left.all,
                     This (Catalogue_Record'Class (Right.all)).all
                  );
            else
               return Object.Less (Left.all, Right.all);
            end if;
         end if;
      end if;
   end "<";

   function "<" (Left, Right : By_Key_Ptr) return Boolean is
   begin
      if Equal (Right, null) then
         return False;
      elsif Equal (Left, null) then
         return True;
      else
         return Left.External < Right.External;
      end if;
   end "<";

   function "=" (Left, Right : By_Name_Ptr) return Boolean is
   begin
      if Equal (Left, null) then
         return Equal (Right, null);
      elsif Equal (Right, null) then
         return False;
      else
         return Left.Name = Right.Name;
      end if;
   end "=";

   function "=" (Left, Right : By_Entity_Ptr) return Boolean is
   begin
      if Equal (Left, null) then
         return Equal (Right, null);
      elsif Equal (Right, null) then
         return False;
      else
         if Left.all in Catalogue_Record'Class then
            if Right.all in Catalogue_Record'Class then
               return
                  Equal
                  (  This (Catalogue_Record'Class (Left.all )).all,
                     This (Catalogue_Record'Class (Right.all)).all
                  );
            else
               return
                  Equal
                  (  This (Catalogue_Record'Class (Left.all)).all,
                     Right.all
                  );
            end if;
         else
            if Right.all in Catalogue_Record'Class then
               return
                  Object.Equal
                  (  Left.all,
                     This (Catalogue_Record'Class (Right.all)).all
                  );
            else
               return Object.Equal (Left.all, Right.all);
            end if;
         end if;
      end if;
   end "=";

   function "=" (Left, Right : By_Key_Ptr) return Boolean is
   begin
      if Equal (Left, null) then
         return Equal (Right, null);
      elsif Equal (Right, null) then
         return False;
      else
         return Left.External = Right.External;
      end if;
   end "=";

   procedure Add
             (  List      : in out Catalogue;
                Object    : Deposit_Ptr;
                External  : Key;
                Name      : Wide_String
             )  is
   begin
      if Name'Length = 0 then
         raise Name_Error;
      end if;
      Create (List, Object, External, Name);
   end Add;

   procedure Add
             (  List     : in out Catalogue;
                Object   : Deposit_Ptr;
                External : Key
             )  is
   begin
      Create (List, Object, External, "");
   end Add;

   procedure Create
             (  List      : in out Catalogue;
                Object    : Deposit_Ptr;
                External  : Key;
                Name      : Wide_String
             )  is
      Index : Integer;
   begin
      Index := Find (List.By_Ptr, By_Entity_Ptr (Object));
      if Index > 0 then
         declare
            Item : Catalogue_Record renames
                      Catalogue_Record (Get (List.By_Ptr, Index).all);
         begin
            if Name /= Item.Name or else External /= Item.External then
               raise Name_Error;            
            end if;
         end;
      else
         declare
            Link : Backward_Link_Ptr :=
                      new Catalogue_Record (Name'Length);
            Item : Catalogue_Record renames Catalogue_Record (Link.all);
         begin
            Item.Use_Count := Item.Use_Count + 1;
            Attach (Link, Object);
            Item.List     := List'Unchecked_Access;
            Item.External := External;
            Item.Name     := Name;
            Add (List.By_Ptr, Item'Unchecked_Access);
            Add (List.By_Key, Item'Unchecked_Access);
            if Name'Length /= 0 then
               Add (List.By_Name, Item'Unchecked_Access);
            end if;
         exception
            when others =>
               Delete (List, Item'Unchecked_Access);
               raise;
         end;
      end if;
   end Create;

   procedure Delete
             (  List : in out Catalogue;
                Item : Catalogue_Record_Ptr
             )  is
   begin
      if Item /= null then
         Remove (List.By_Ptr,  By_Entity_Ptr (Item));
         Remove (List.By_Key,  By_Key_Ptr    (Item));
         Remove (List.By_Name, By_Name_Ptr   (Item));
         Detach (Item.all);
         case Item.Use_Count is
            when 0 =>
               raise Program_Error;
            when 1 =>
               Item.Use_Count := 0;
               declare
                  Ptr : Catalogue_Record_Ptr := Item;
               begin
                  Free (Ptr);
               end;
            when others =>
               Item.Use_Count := Item.Use_Count - 1;
         end case;
      end if;
   end Delete;

   procedure Delete (List : in out Catalogue; ID : Deposit_Ptr) is
      Index : Integer := Find (List, ID);
   begin
      if Index > 0 then
         Delete
         (  List,
            Catalogue_Record
            (  Get (List.By_Ptr, Index).all
            ) 'Unchecked_Access
         );
      end if;
   end Delete;

   procedure Delete (List : in out Catalogue; ID : Wide_String) is
      Index : Integer := Find (List, ID);
   begin
      if Index > 0 then
         Delete
         (  List,
            Catalogue_Record_Ptr (Get (List.By_Name, Index))
         );
      end if;
   end Delete;

   procedure Delete (List : in out Catalogue; ID : Key) is 
      Index : Integer := Find (List, ID);
   begin
      if Index > 0 then
         Delete
         (  List,
            Catalogue_Record_Ptr (Get (List.By_Key, Index))
         );
      end if;
   end Delete;

   procedure Deleted
             (  Item  : in out Catalogue_Record;
                Temps : in out Deposit_Container'Class
             )  is
      Links : Key_Sets.Set;
      Mutex : Read_Mutex (Item.List.Storage);
   begin
      for Dependant in Positive'Range loop
         Add
         (  Links,
            Key
            (  Get_Dependant
               (  Item.List.Storage,
                  Item.External,
                  Dependant
        )  )  );
      end loop;
      Commit (Mutex);
   exception
      when Constraint_Error =>
         raise Ada.IO_Exceptions.Data_Error;
      when Ada.IO_Exceptions.End_Error =>
         declare
            Unloaded : Deposit_Handle;
         begin
            for Dependant in 1..Get_Size (Links) loop
               if not Is_In (Item.List.all, Get (Links, Dependant)) then
                  Unloaded :=
                     Get
                     (  Item.List.Storage,
                        Get (Links, Dependant)
                     );
                  Add (Temps, Ptr (Unloaded));
               end if;
            end loop;
         end;         
         Commit (Mutex);
   end Deleted;

   procedure Destroyed (Item : in out Catalogue_Record) is
      Object : Deposit'Class renames This (Item).all;
   begin
      if Item.Name_Length = 0 then
         declare
            Mutex : Write_Mutex (Item.List.Storage);
         begin
            if Has_Dependants (Item.List.Storage, Item.External) then
               if Is_Modified (Object) then
                  Put
                  (  Item.List.Storage.all,
                     Item.External,
                     Object
                  );
                  Reset_Modified (Object);
               end if;
            else
               Delete (Item.List.Storage.all, Item.External);
            end if;
            Commit (Mutex);
         end;
      elsif Is_Modified (Object) then
         declare
            Mutex : Write_Mutex (Item.List.Storage);
         begin
            Put
            (  Item.List.Storage.all,
               Item.External,
               Object
            );
            Commit (Mutex);
         end;
         Reset_Modified (Object);
      end if;
      Delete (Item.List.all, Item'Unchecked_Access);
   end Destroyed;

   procedure Erase (List : in out Catalogue) is
   begin
      while not Is_Empty (List.By_Key) loop
         Delete
         (  List,
            Catalogue_Record_Ptr (Get (List.By_Key, 1))
         );
      end loop;
   end Erase;

   procedure Finalize (List : in out Catalogue) is
   begin
      Erase (List);
   end Finalize;

   function Find (List : Catalogue; ID : Wide_String)
      return Integer is
      Item : aliased Catalogue_Record (ID'Length);
   begin
      Item.Name := ID;
      return Find (List.By_Name, Item'Unchecked_Access);
   end Find;

   function Find (List : Catalogue; ID : Deposit_Ptr) return Integer is
   begin
      return Find (List.By_Ptr, ID.all'Unchecked_Access);
   end Find;

   function Find (List : Catalogue; ID : Key) return Integer is
      Item : aliased Catalogue_Record (0);
   begin
      Item.External := ID;
      return Find (List.By_Key, Item'Unchecked_Access);
   end Find;

   function Get (List : Catalogue; ID : Wide_String) return Key is
      Index : Integer := Find (List, ID);
   begin
      if Index > 0 then
         return Get (List.By_Name, Index).External;
      end if;
      raise End_Error;
   end Get;

   function Get (List : Catalogue; ID : Wide_String)
      return Deposit_Ptr is
      Index : Integer := Find (List, ID);
   begin
      if Index > 0 then
         return This (Get (List.By_Name, Index).all);
      end if;
      raise End_Error;
   end Get;

   function Get (List : Catalogue; ID : Deposit_Ptr) return Key is
      Index : Integer := Find (List, ID);
   begin
      if Index > 0 then
         return
            Catalogue_Record
            (  Get (List.By_Ptr, Index).all
            ) .External;
      end if;
      raise End_Error;
   end Get;

   function Get (List : Catalogue; ID : Deposit_Ptr)
      return Wide_String is
      Index : Integer := Find (List, ID);
   begin
      if Index > 0 then
         declare
            Item : Catalogue_Record renames
                      Catalogue_Record (Get (List.By_Ptr, Index).all);
         begin
            if Item.Name'Length > 0 then
               return Item.Name;
            end if;
         end;
      end if;
      raise End_Error;
   end Get;

   function Get (List : Catalogue; ID : Key) return Wide_String is
      Index : Integer := Find (List, ID);
   begin
      if (  Index > 0
         and then
            Get (List.By_Key, Index).Name'Length > 0
         )
      then
         return Get (List.By_Key, Index).Name;
      end if;
      raise End_Error;
   end Get;

   function Get (List : Catalogue; ID : Key) return Deposit_Ptr is
      Index : Integer := Find (List, ID);
   begin
      if Index > 0 then
         return This (Get (List.By_Key, Index).all);
      end if;
      raise End_Error;
   end Get;

   function Is_In (List : Catalogue; ID : Wide_String)
      return Boolean is
   begin
      return Find (List, ID) > 0;
   end Is_In;

   function Is_In (List : Catalogue; ID : Key)
      return Boolean is
   begin
      return Find (List, ID) > 0;
   end Is_In;

   function Is_In (List : Catalogue; ID : Deposit_Ptr)
      return Boolean is
   begin
      return Find (List, ID) > 0;
   end Is_In;

   procedure Rename
             (  List : in out Catalogue;
                Old  : Catalogue_Record_Ptr;
                Name : Wide_String := ""
             )  is
      Link : Backward_Link_Ptr;
   begin
      if Old.Name = Name then
         return;
      end if;
      if Is_In (List, Name) then
         raise Name_Error;
      end if;
      Link := new Catalogue_Record (Name'Length);
      declare
         Item : Catalogue_Record renames Catalogue_Record (Link.all);
      begin
         Item.Use_Count := Item.Use_Count + 1;
         Item.List      := Old.List;
         Item.External  := Old.External;
         Item.Name      := Name;
         Attach (Link, This (Old.all));
         Delete (List, Old);
         Add (List.By_Ptr, Item'Unchecked_Access);
         Add (List.By_Key, Item'Unchecked_Access);
         if Name'Length /= 0 then
            Add (List.By_Name, Item'Unchecked_Access);
         end if;
      exception
         when others =>
            Delete (List, Item'Unchecked_Access);
            raise;
      end;
   end Rename;

   procedure Rename
             (  List : in out Catalogue;
                ID   : Wide_String;
                Name : Wide_String
             )  is
      Index : Integer := Find (List, ID);
   begin
      if Index <= 0 then
         raise End_Error;
      end if;
      Rename
      (  List,
         Catalogue_Record_Ptr (Get (List.By_Name, Index)),
         Name
      );
   end Rename;

   procedure Rename
             (  List : in out Catalogue;
                ID   : Deposit_Ptr;
                Name : Wide_String
             )  is
      Index : Integer := Find (List, ID);
   begin
      if Index <= 0 then
         raise End_Error;
      end if;
      Rename
      (  List,
         Catalogue_Record
         (  Get (List.By_Ptr, Index).all
         ) 'Unchecked_Access,
         Name
      );
   end Rename;

   procedure Rename
             (  List : in out Catalogue;
                ID   : Key;
                Name : Wide_String
             )  is
      Index : Integer := Find (List, ID);
   begin
      if Index <= 0 then
         raise End_Error;
      end if;
      Rename
      (  List,
         Catalogue_Record_Ptr (Get (List.By_Key, Index)),
         Name
      );
   end Rename;

   procedure Unname (List : in out Catalogue; ID : Wide_String) is
      Index : Integer := Find (List, ID);
   begin
      if Index <= 0 then
         raise End_Error;
      end if;
      Rename
      (  List,
         Catalogue_Record_Ptr (Get (List.By_Name, Index))
      );
   end Unname;

   procedure Unname (List : in out Catalogue; ID : Deposit_Ptr) is
      Index : Integer := Find (List, ID);
   begin
      if Index <= 0 then
         raise End_Error;
      end if;
      Rename
      (  List,
         Catalogue_Record
         (  Get (List.By_Ptr, Index).all
         ) 'Unchecked_Access
      );
   end Unname;
 
   procedure Unname (List : in out Catalogue; ID : Key) is
      Index : Integer := Find (List, ID);
   begin
      if Index <= 0 then
         raise End_Error;
      end if;
      Rename
      (  List,
         Catalogue_Record_Ptr (Get (List.By_Key, Index))
      );
   end Unname;

end Persistent.Data_Bank.Index;