--                                                                    --
--  package                         Copyright (c)  Dmitry A. Kazakov  --
--     Persistent.Data_Bank.Index                  Luebeck            --
--  Interface                                      Autumn, 2004       --
--                                                                    --
--                                Last revision :  12:32 10 Jun 2003  --
--                                                                    --
--  This  library  is  free software; you can redistribute it and/or  --
--  modify it under the terms of the GNU General Public  License  as  --
--  published by the Free Software Foundation; either version  2  of  --
--  the License, or (at your option) any later version. This library  --
--  is distributed in the hope that it will be useful,  but  WITHOUT  --
--  ANY   WARRANTY;   without   even   the   implied   warranty   of  --
--  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU  --
--  General  Public  License  for  more  details.  You  should  have  --
--  received  a  copy  of  the GNU General Public License along with  --
--  this library; if not, write to  the  Free  Software  Foundation,  --
--  Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.    --
--                                                                    --
--  As a special exception, if other files instantiate generics from  --
--  this unit, or you link this unit with other files to produce  an  --
--  executable, this unit does not by  itself  cause  the  resulting  --
--  executable to be covered by the GNU General Public License. This  --
--  exception  does not however invalidate any other reasons why the  --
--  executable file might be covered by the GNU Public License.       --
--____________________________________________________________________--
--
--  This  package  provides an index for a persistent storage. The index
--  can  be  searched  by  object's  name,  pointer, key. The package is
--  generic. The generic parameters are:
--
--  (o)  Data_Bank is the base type of the persistent  storage  objects.
--       It can be any descendant of Data_Bank_Object.
--
--  (o)  Key  is  a  descendant  of  Persistent_Key  to be used with the
--       persistent storage.
--
generic
   type Data_Bank is abstract new Data_Bank_Object with private;
   type Key is new Persistent_Key with private;
package Persistent.Data_Bank.Index is
   type Catalogue (Storage : access Data_Bank'Class) is
      new Ada.Finalization.Limited_Controlled with private;
   type Catalogue_Ptr is access all Catalogue;
--
-- Add - A new record to the Catalogue
--
--    List      - Of the bound handles
--    External  - The external key
--    Internal  - The internal key
--  [ Name ]    - The name
--
-- When Name is specified but empty, Name_Error is propagated.
--
-- Exceptions :
--
--    Name_Error - Name or handle conflict
--
   procedure Add
             (  List     : in out Catalogue;
                Object   : Deposit_Ptr;
                External : Key;
                Name     : Wide_String
             );
   procedure Add
             (  List     : in out Catalogue;
                Object   : Deposit_Ptr;
                External : Key
             );
--
-- Delete - Delete from the Catalogue
--
--    List - Of the bound handles
--    ID   - Of the record
--
-- Nothing happens if it is not there
--
   procedure Delete (List : in out Catalogue; ID : Wide_String);
   procedure Delete (List : in out Catalogue; ID : Deposit_Ptr);
   procedure Delete (List : in out Catalogue; ID : Key);
--
-- Erase -- Deletes all index records
--
--    List - Of the bound handles
--
   procedure Erase (List : in out Catalogue);
--
-- Finalize -- Destructor
--
--    List - Of the bound handles
--
   procedure Finalize (List : in out Catalogue);
--
-- Get -- From the Catalogue
--
--    List - Of the bound handles
--    ID   - To search for
--
-- Returns :
--
--    The result
--
-- Exceptions :
--
--    End_Error - Nothing found
--
   function Get (List : Catalogue; ID : Deposit_Ptr) return Key;
   function Get (List : Catalogue; ID : Deposit_Ptr) return Wide_String;
   function Get (List : Catalogue; ID : Key        ) return Deposit_Ptr;
   function Get (List : Catalogue; ID : Key        ) return Wide_String;
   function Get (List : Catalogue; ID : Wide_String) return Deposit_Ptr;
   function Get (List : Catalogue; ID : Wide_String) return Key;
   pragma Inline (Get);
--
-- Is_In -- Check if in the Catalogue
--
--    List - Of the bound handles
--    ID   - To search for
--
-- Returns :
--
--    True if is in
--
   function Is_In (List : Catalogue; ID : Wide_String)
      return Boolean;
   function Is_In (List : Catalogue; ID : Key)
      return Boolean;
   function Is_In (List : Catalogue; ID : Deposit_Ptr)
      return Boolean;
--
-- Rename - In the Catalogue
--
--    List - Of the bound handles
--    ID   - Of the record to rename
--    Name - The new name
--
-- Exceptions :
--
--    End_Error  - No such record
--    Name_Error - Name conflict
--
   procedure Rename
             (  List : in out Catalogue;
                ID   : Wide_String;
                Name : Wide_String
             );
   procedure Rename
             (  List : in out Catalogue;
                ID   : Deposit_Ptr;
                Name : Wide_String
             );
   procedure Rename
             (  List : in out Catalogue;
                ID   : Key;
                Name : Wide_String
             );
--
-- Unname - In the Catalogue
--
--    List - Of the bound handles
--    ID   - Of the record to rename
--
-- Exceptions :
--
--    End_Error - No such record
--
   procedure Unname (List : in out Catalogue; ID : Wide_String);
   procedure Unname (List : in out Catalogue; ID : Deposit_Ptr); 
   procedure Unname (List : in out Catalogue; ID : Key);

private
--
-- Catalogue_Record -- Bound object record
--
--    Name_Length - The object's name length
--    List        - The Catalogue of bound objects
--    Key         - The object's external key
--    Object      - A pointer to the object
--    Note        - A pointer to the notification used for the object
--    Name        - The object's name
--
   type Catalogue_Record (Name_Length : Natural) is
      new Backward_Link with
   record
      List     : Catalogue_Ptr;
      External : Key;
      Name     : Wide_String (1..Name_Length);
   end record;
   type Catalogue_Record_Ptr is access all Catalogue_Record;
--
-- Deleted -- Overrides Object.Archived...
--
   procedure Deleted
             (  Item  : in out Catalogue_Record;
                Temps : in out Deposit_Container'Class
             );
--
-- Destroyed -- Overrides Object.Archived...
--
   procedure Destroyed (Item : in out Catalogue_Record);

   type By_Name_Ptr is new Catalogue_Record_Ptr;
   function Equal (Left, Right : By_Name_Ptr)
      return Boolean renames "=";
   function "<" (Left, Right : By_Name_Ptr) return Boolean;
   function "=" (Left, Right : By_Name_Ptr) return Boolean;

   type By_Entity_Ptr is access all Object.Entity'Class;
   function Equal (Left, Right : By_Entity_Ptr)
      return Boolean renames "=";
   function "<" (Left, Right : By_Entity_Ptr) return Boolean;
   function "=" (Left, Right : By_Entity_Ptr) return Boolean;

   type By_Key_Ptr is new Catalogue_Record_Ptr;
   function Equal (Left, Right : By_Key_Ptr)
      return Boolean renames "=";
   function "<" (Left, Right : By_Key_Ptr) return Boolean;
   function "=" (Left, Right : By_Key_Ptr) return Boolean;

   package Name_Map is new Generic_Set (By_Name_Ptr,   null);
   package Ptr_Map  is new Generic_Set (By_Entity_Ptr, null);
   package Key_Map  is new Generic_Set (By_Key_Ptr,    null);

   type Catalogue
        (  Storage : access Data_Bank'Class
        )  is new Ada.Finalization.Limited_Controlled with
   record
      By_Name : Name_Map.Set;
      By_Ptr  : Ptr_Map.Set;
      By_Key  : Key_Map.Set;
   end record;
--
-- Find -- Search the Catalogue (a map of)
--
--    List - To be searched
--    ID   - To search for
--
-- Returns :
--
--    Catalogue in the corresponding map
--
   function Find (List : Catalogue; ID : Wide_String) return Integer;
   function Find (List : Catalogue; ID : Deposit_Ptr) return Integer;
   function Find (List : Catalogue; ID : Key)         return Integer;
--
-- Key_Set -- Sets of Keys
--
   package Key_Sets is new Generic_Set (Key, Null_Key);
   use Key_Sets;
       
end Persistent.Data_Bank.Index;