--                                                                    --
--  package                         Copyright (c)  Dmitry A. Kazakov  --
--     Persistent.Data_Bank.Indexed                Luebeck            --
--  Implementation                                 Autumn, 2004       --
--                                                                    --
--                                Last revision :  12:32 10 Jun 2003  --
--                                                                    --
--  This  library  is  free software; you can redistribute it and/or  --
--  modify it under the terms of the GNU General Public  License  as  --
--  published by the Free Software Foundation; either version  2  of  --
--  the License, or (at your option) any later version. This library  --
--  is distributed in the hope that it will be useful,  but  WITHOUT  --
--  ANY   WARRANTY;   without   even   the   implied   warranty   of  --
--  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU  --
--  General  Public  License  for  more  details.  You  should  have  --
--  received  a  copy  of  the GNU General Public License along with  --
--  this library; if not, write to  the  Free  Software  Foundation,  --
--  Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.    --
--                                                                    --
--  As a special exception, if other files instantiate generics from  --
--  this unit, or you link this unit with other files to produce  an  --
--  executable, this unit does not by  itself  cause  the  resulting  --
--  executable to be covered by the GNU General Public License. This  --
--  exception  does not however invalidate any other reasons why the  --
--  executable file might be covered by the GNU Public License.       --
--____________________________________________________________________--

with Ada.Characters.Handling;        use Ada.Characters.Handling;
with Ada.Exceptions;                 use Ada.Exceptions;
with Ada.IO_Exceptions;              use Ada.IO_Exceptions;
with Object.Archived.Lists;          use Object.Archived.Lists;
with Persistent.Data_Bank.Reference; use Persistent.Data_Bank.Reference;
with Persistent.Handle;              use Persistent.Handle;
with Strings_Edit;                   use Strings_Edit;
with Strings_Edit.Integers;          use Strings_Edit.Integers;
with Strings_Edit.Quoted;            use Strings_Edit.Quoted;

package body Persistent.Data_Bank.Indexed is
   use Storage_Index;

   Backward_Link_Quote : constant Character := '?';
   Direct_Link_Quote   : constant Character := '/';
--
-- Restore_Object -- Recursively restore object by its key
--
--    Storage    - Persistent data storage (a pointer to)
--    Key        - The key of the object being restored
--
-- Returns :
--
--    Pointer to the newly created object
--
   function Restore_By_Key
            (  Storage : access Indexed_Storage_Object'Class;
               Key     : Persistent_Key'Class
            )  return Deposit_Ptr;
--
-- Restore_Object -- Recursively restore object
--
--    Storage    - Persistent data storage (a pointer to)
--    Key        - The key of the object being restored
--    Class      - The object class
--    Data       - The object description
--    Parameters - Additional parameters
--    Object     - The result
--
   procedure Restore_Object
             (  Storage    : in out Indexed_Storage_Object'Class;
                Key        : Persistent_Key'Class;
                Class      : String;
                Data       : String;
                Parameters : String;
                Object     : out Deposit_Ptr
             );
--
-- Store_Dependencies
--
--    Storage        - Persistent data storage
--    Data           - To write
--    Parameters     - To write
--    Object         - The object
--    Direct_Links   - Of the object (output)
--    Backward_Links - Of the object (output)
--
   procedure Store_Dependencies
             (  Storage        : in out Indexed_Storage_Object'Class;
                Data           : in out Unbounded_String;
                Parameters     : in out Unbounded_String; 
                Object         : Deposit'Class;
                Direct_Links   : in out Deposit_Set;
                Backward_Links : in out Deposit_Set
             );
--
-- Unchecked_Is_In -- Is_In under a mutex taken by the caller
--
   function Unchecked_Is_In
            (  Storage : Indexed_Storage_Object;
               Name    : Wide_String
            )  return Boolean;

   procedure Add_To_Index
             (  Storage : in out Indexed_Storage_Object;
                Object  : in out Deposit_Handles.Handle
             )  is
   begin
      if not Is_Valid (Object) then
         raise Constraint_Error;
      end if;
      if not Is_In (Storage.Index, Ptr (Object)) then
         declare
            Direct_Links   : Deposit_Set;
            Backward_Links : Deposit_Set;
            Data           : Unbounded_String;
            Parameters     : Unbounded_String;
         begin
            Store_Dependencies
            (  Storage        => Storage,
               Data           => Data,
               Parameters     => Parameters,
               Object         => Ptr (Object).all,
               Direct_Links   => Direct_Links,
               Backward_Links => Backward_Links
            );
            Add
            (  List     => Storage.Index,
               Object   => Ptr (Object),
               External =>
                  Key
                  (  Store
                     (  Storage        => Storage.Index.Storage,
                        Class          => Get_Class (Object),
                        Data           => To_String (Data),
                        Parameters     => To_String (Parameters),
                        Direct_Links   => Direct_Links,
                        Backward_Links => Backward_Links
            )    )  );
         end;
      end if;
   end Add_To_Index;

   procedure Create
             (  Storage        : in out Indexed_Storage_Object;
                Object         : Deposit_Handles.Handle;
                ID             : out Key;
                Name           : Wide_String;
                Class          : String;
                Data           : String;
                Parameters     : String;
                Direct_Links   : Deposit_Set;
                Backward_Links : Deposit_Set              
            )   is
   begin
      if not Is_Valid (Object) then
         raise Constraint_Error;
      end if;
      ID :=
         Key
         (  Store
            (  Storage        => Storage.Index.Storage,
               Name           => Name,
               Class          => Class,
               Data           => Data,
               Parameters     => Parameters,
               Direct_Links   => Direct_Links,
               Backward_Links => Backward_Links
         )  );
      Add (Storage.Index, Ptr (Object), ID, Name);
   end Create;

   function Get
            (  Storage : access Indexed_Storage_Object;
               Name    : Wide_String
            )  return Deposit_Handles.Handle is
      Object : Deposit_Ptr;
   begin
      begin
         Object := Get (Storage.Index, Name);
      exception
         when Ada.IO_Exceptions.End_Error =>
            declare
               Mutex : Read_Mutex (Storage.Index.Storage);
               Storage_Key : Key renames
                  Key (Find (Storage.Index.Storage, Name));
            begin
               Object := Restore_By_Key (Storage, Storage_Key);
               Add (Storage.Index, Object, Storage_Key, Name);
               Commit (Mutex);
            end;
      end;
      return Ref (Object);
   exception
      when others =>
         Free (Object);
         raise;
   end Get;

   function Get
            (  Storage : access Indexed_Storage_Object;
               Key     : Persistent_Key'Class
            )  return Deposit_Handles.Handle is
   begin
      if Key not in Persistent.Data_Bank.Indexed.Key'Class then
         raise Ada.IO_Exceptions.End_Error;
      end if;
      declare
         Object      : Deposit_Ptr;
         Storage_Key : Persistent.Data_Bank.Indexed.Key renames
                          Persistent.Data_Bank.Indexed.Key (Key);
      begin
         begin
            Object := Get (Storage.Index, Storage_Key);
         exception
            when Ada.IO_Exceptions.End_Error =>
               Object := Restore_By_Key (Storage, Storage_Key);
               begin
                  --
                  -- Get the object's name
                  --
                  Add
                  (  Storage.Index,
                     Object,
                     Storage_Key,
                     Get_Name (Storage.Index.Storage, Key)
                  );
               exception
                  when Ada.IO_Exceptions.Name_Error =>
                     --
                     -- The object is anonymous
                     --
                     Add (Storage.Index, Object, Storage_Key);
               end;
         end;
         return Ref (Object);
      exception
         when others =>
            Free (Object);
            raise;
      end;
   end Get;

   function Get_Key
            (  Storage : Indexed_Storage_Object;
               Object  : Deposit_Ptr
            )  return Key is
   begin
      return Get (Storage.Index, Object);
   exception
      when Constraint_Error =>
         raise;
      when Ada.IO_Exceptions.End_Error =>
         raise Constraint_Error;    
   end Get_Key;

   function Get_Name
            (  Storage : access Indexed_Storage_Object;
               Object  : Deposit_Handles.Handle
            )  return Wide_String is
   begin
      if Is_Valid (Object) then
         declare
            Result : Wide_String := Get (Storage.Index, Ptr (Object));
         begin
            if Result'Length > 0 then
               return Result;
            end if;
            raise Ada.IO_Exceptions.Name_Error;
         exception
            when Ada.IO_Exceptions.End_Error =>
               null;
         end;
      end if;
      raise Constraint_Error;
   end Get_Name;

   function Is_In
            (  Storage : access Indexed_Storage_Object;
               Name    : Wide_String
            )  return Boolean is
   begin
      if Is_In (Storage.Index, Name) then
         return True;
      end if;
      declare
         Mutex : Read_Mutex (Storage.Index.Storage);
         Key   : Persistent_Key'Class :=
                    Find (Storage.Index.Storage, Name);
      begin
         Commit (Mutex);
         return True;
      end;
   exception
      when Ada.IO_Exceptions.End_Error =>
         return False;
   end Is_In;

   function Is_In
            (  Storage : access Indexed_Storage_Object;
               Object  : Deposit_Handles.Handle
            )  return Boolean is
   begin
      return Is_In (Storage.Index, Ptr (Object));
   end Is_In;

   function Is_Indexed
            (  Storage : access Indexed_Storage_Object;
               Name    : Wide_String
            )  return Boolean is
   begin
      return Is_In (Storage.Index, Name);
   end Is_Indexed;

   procedure Purge (Storage : in out Indexed_Storage_Object'Class) is
   begin
      Erase (Storage.Index);
   end Purge;

   procedure Put
             (  Storage : in out Indexed_Storage_Object;
                Object  : in out Deposit_Handles.Handle;
                Name    : Wide_String
             )  is
      Direct_Links   : Deposit_Set;
      Backward_Links : Deposit_Set;
   begin
      if not Is_Valid (Object) then
         raise Constraint_Error;
      end if;
      if Is_In (Storage.Index, Ptr (Object)) then
         --
         -- The object is already in the data base, the only thing to do
         -- is to check that the name is same. 
         --
         declare
            Mutex : Read_Mutex (Storage.Index.Storage);
         begin
            if Get (Storage.Index, Ptr (Object)) /= Name then
               raise Ada.IO_Exceptions.Name_Error;
            end if;
            Commit (Mutex);
         end;
      else
         --
         -- The  object  is not there. We have to check that the name is
         -- free and then store it.
         --
         declare
            Mutex : Write_Mutex (Storage.Index.Storage);
         begin
            if Unchecked_Is_In (Storage, Name) then
               Raise_Exception
               (  Ada.IO_Exceptions.Name_Error'Identity,
                  (  "Another object '"
                  &  To_String (Name, '?')
                  &  "' already exists"
               )  );
            end if;
            declare
               Data       : Unbounded_String;
               Parameters : Unbounded_String;
            begin
               Store_Dependencies
               (  Storage        => Storage,
                  Data           => Data,
                  Parameters     => Parameters,
                  Object         => Ptr (Object).all,
                  Direct_Links   => Direct_Links,
                  Backward_Links => Backward_Links
               );
               Add
               (  List     => Storage.Index,
                  Object   => Ptr (Object),
                  Name     => Name,
                  External =>
                     Key
                     (  Store
                        (  Storage        => Storage.Index.Storage,
                           Name           => Name,
                           Class          => Get_Class (Object),
                           Data           => To_String (Data),
                           Parameters     => To_String (Parameters),
                           Direct_Links   => Direct_Links,
                           Backward_Links => Backward_Links
               )    )  );
            end;
            Commit (Mutex); 
         end;
      end if;
   end Put;

   procedure Put
             (  Storage : in out Indexed_Storage_Object;
                Object  : in out Deposit_Handles.Handle
             )  is
      Mutex : Write_Mutex (Storage.Index.Storage);
   begin
      Add_To_Index (Storage, Object);
      Commit (Mutex);
   end Put;

   procedure Put
             (  Storage : in out Indexed_Storage_Object;
                Key     : Persistent_Key'Class;
                Object  : Deposit'Class
             )  is
      Direct_Links   : Deposit_Set;
      Backward_Links : Deposit_Set;
      Data           : Unbounded_String;
      Parameters     : Unbounded_String;
   begin
      Store_Dependencies
      (  Storage        => Storage,
         Data           => Data,
         Parameters     => Parameters,
         Object         => Object,
         Direct_Links   => Direct_Links,
         Backward_Links => Backward_Links
      );
      Update
      (  Storage        => Storage.Index.Storage.all,
         Key            => Key,
         Class          => Get_Class (Object),
         Data           => To_String (Data),
         Parameters     => To_String (Parameters),
         Direct_Links   => Direct_Links,
         Backward_Links => Backward_Links
      );
   end Put;

   procedure Rename
             (  Storage  : in out Indexed_Storage_Object;
                Object   : Wide_String;
                New_Name : Wide_String
             )  is
      Mutex : Write_Mutex (Storage.Index.Storage);
   begin
      if Is_In (Storage.Index, Object) then
         --
         -- Object  is  bound,  so  we  get its handle and use rename by
         -- handle. 
         --
         declare
            This : Deposit_Handles.Handle :=
                      Get (Storage.Index.Storage, Object);
         begin
            Rename (Storage.Index.Storage.all, This, New_Name);
         end;
      else
         --
         -- Object is not bound. We rename it in the storage
         --
         declare
            Key : Persistent_Key'Class :=
                     Find (Storage.Index.Storage, Object);
         begin
            Rename (Storage.Index.Storage.all, Key, New_Name); 
         end;
      end if;
      Commit (Mutex);
   end Rename;

   procedure Rename
             (  Storage  : in out Indexed_Storage_Object;
                Object   : in out Deposit_Handles.Handle;
                New_Name : Wide_String
             )  is
   begin
      if not Is_In (Storage.Index, Ptr (Object)) then
         Raise_Exception
         (  Constraint_Error'Identity,
            "Renaming a non-persistent object"
         );
      end if;
      if New_Name'Length < 0 then
         Raise_Exception
         (  Ada.IO_Exceptions.Name_Error'Identity,
            "No empty names allowed"
         );
      end if;
      declare
         Mutex : Write_Mutex (Storage.Index.Storage);
      begin
         if Unchecked_Is_In (Storage, New_Name) then
            if Get (Storage.Index, Ptr (Object)) /= New_Name then
               Raise_Exception
               (  Ada.IO_Exceptions.Name_Error'Identity,
                  (  "Object '"
                  &  To_String (New_Name, '?')
                  &  "' already exists"
               )  );
            end if;
         else
            Rename
            (  Storage.Index.Storage.all,
               Key'(Get (Storage.Index, Ptr (Object))),
               New_Name
            );
            Rename (Storage.Index, Ptr (Object), New_Name);
         end if;
         Commit (Mutex);
      end;
   end Rename;

   function Restore_By_Key
            (  Storage : access Indexed_Storage_Object'Class;
               Key     : Persistent_Key'Class
            )  return Deposit_Ptr is
      Class      : Unbounded_String;
      Data       : Unbounded_String;
      Parameters : Unbounded_String;
      Object     : Deposit_Ptr;
   begin
      Get_Data (Storage.all, Key, Class, Data, Parameters);
      Restore_Object
      (  Storage.all,
         Key,
         To_String (Class),
         To_String (Data),
         To_String (Parameters),
         Object
      );
      return Object;
   end Restore_By_Key;

   procedure Restore_Object
             (  Storage    : in out Indexed_Storage_Object'Class;
                Key        : Persistent_Key'Class;
                Class      : String;
                Data       : String;
                Parameters : String;
                Object     : out Deposit_Ptr
             )  is
      Referent : Deposit_Handles.Handle;
      List     : Deposit_List;
      Pointer  : aliased Integer := Parameters'First;
   begin
      loop
         Get (Parameters, Pointer);
         exit when Pointer > Parameters'Last;
         case Parameters (Pointer) is
            when '@' =>    -- Reference to the data base
               Pointer := Pointer + 1;
               Add
               (  List,
                  Ref
                  (  To_Persistent_Storage_Ptr
                     (  Storage'Unchecked_Access
                  )  ),
                  Key
               ); 
            when Direct_Link_Quote =>       -- Key of (direct link)
               Referent :=
                  Get
                  (  Storage'Unchecked_Access, 
                     Value
                     (  Storage,
                        Get_Quoted
                        (  Parameters,
                           Pointer'Unchecked_Access,
                           Direct_Link_Quote
                  )  )  );
               Add (List, Ptr (Referent), False);
            when Backward_Link_Quote =>    -- Key of (backward link)
               Referent :=
                  Get
                  (  Storage'Unchecked_Access, 
                     Value
                     (  Storage,
                        Get_Quoted
                        (  Parameters,
                           Pointer'Unchecked_Access,
                           Backward_Link_Quote
                  )  )  );
               Add (List, Ptr (Referent), True);
            when others => -- Number of an object in the list to use
               declare
                  Position : Integer;
               begin
                  Get (Parameters, Pointer, Position);
                  Add (List, Get (List, Position), False);
               exception
                  when others =>
                     raise Ada.IO_Exceptions.Data_Error;
               end;
         end case;
      end loop;        
      Pointer := Data'First;
      Get (Data, Pointer);
      Create
      (  Source  => Data,
         Pointer => Pointer,
         Class   => Class,
         List    => List,
         Object  => Object
      );
      Get (Data, Pointer);
      if Pointer /= Data'Last + 1 then
         raise Ada.IO_Exceptions.Data_Error;
      end if;
   exception
      when others =>
         Free (Object);
         raise;
   end Restore_Object;

   procedure Store_Dependencies
             (  Storage        : in out Indexed_Storage_Object'Class;
                Data           : in out Unbounded_String;
                Parameters     : in out Unbounded_String;
                Object         : Deposit'Class;
                Direct_Links   : in out Deposit_Set;
                Backward_Links : in out Deposit_Set
             )  is
      Referent_Key : Key;
      Mark         : Character;
      Item         : Deposit_Ptr;
      List         : Deposit_List;
   begin
      Get_Referents (Object, List);
      for Position in 1..Get_Size (List) loop
         if Position > 1 then
            Append (Parameters, ' ');
         end if;
         Item := Get (List, Position);
         if Item.all in Self_Reference'Class then
            Append (Parameters, '@');
         else
            if Is_Backward (List, Item) then
               Mark := Backward_Link_Quote;
            else
               Mark := Direct_Link_Quote;
            end if;
            if Is_In (Storage.Index, Item) then
               Append
               (  Parameters,
                  Quote
                  (  Image (Storage, Get (Storage.Index, Item)),
                     Mark
               )  ); 
            else
               declare
                  Referent_Param : Unbounded_String;
                  Referent_Data  : Unbounded_String;
               begin
                  Store_Dependencies
                  (  Storage        => Storage,
                     Data           => Referent_Data,
                     Parameters     => Referent_Param,
                     Object         => Item.all,
                     Direct_Links   => Direct_Links,
                     Backward_Links => Backward_Links
                  );
                  Referent_Key :=
                     Key
                     (  Store
                        (  Storage        => Storage.Index.Storage,
                           Class          => Get_Class (Item.all),
                           Data           => To_String (Referent_Data),
                           Parameters     => To_String (Referent_Param),
                           Direct_Links   => Direct_Links,
                           Backward_Links => Backward_Links
                     )  );
               end;
               Add (Storage.Index, Item, Referent_Key);
               Append
               (  Parameters,
                  Quote (Image (Storage, Referent_Key), Mark)
               );
            end if;
         end if;
      end loop;
      if (  Get_Size (List) = 0
         or else
            Get (List, 1).all not in Self_Reference'Class
         )
      then
         declare
            Size : Positive := 1024;
         begin
            loop
               declare
                  Buffer  : String (1..Size);
                  Pointer : Positive := Buffer'First;
               begin
                  Store (Buffer, Pointer, Object);
                  Append (Data, Buffer (Buffer'First..Pointer - 1));
                  exit;
               exception
                  when Ada.IO_Exceptions.Layout_Error =>
                     Size := Size * 2;
               end;
            end loop;
         end;
      end if;
      Erase (Direct_Links);
      Erase (Backward_Links);
      for Position in 1..Get_Size (List) loop
         Item := Get (List, Position);
         if Item.all not in Self_Reference'Class then
            if Is_Backward (List, Item) then
               Add (Backward_Links, Item);
            else
               Add (Direct_Links, Item);
            end if;
         end if;
      end loop;
   end Store_Dependencies;

   function Unchecked_Is_In
            (  Storage : Indexed_Storage_Object;
               Name    : Wide_String
            )  return Boolean is
   begin
      if Is_In (Storage.Index, Name) then
         return True;
      end if;
      declare
         Key : Persistent_Key'Class :=
                  Find (Storage.Index.Storage, Name);
      begin
         return True;
      end;
   exception
      when Ada.IO_Exceptions.End_Error =>
         return False;
   end Unchecked_Is_In;

   procedure Unname
             (  Storage : in out Indexed_Storage_Object;
                Object  : Wide_String
             )  is
      Mutex : Write_Mutex (Storage.Index.Storage);
   begin
      if Is_In (Storage.Index, Object) then
         --
         -- Object is bound
         --
         declare
            ID : Key := Get (Storage.Index, Object);
         begin
            Unname (Storage.Index, ID); 
            Unname (Storage.Index.Storage.all, ID);
         end;
      else
         --
         -- Object is not bound. We unname it in the storage
         --
         declare
            Key : Persistent_Key'Class :=
                     Find (Storage.Index.Storage, Object);
         begin
            Unname (Storage.Index.Storage.all, Key); 
         exception
            when Ada.IO_Exceptions.End_Error =>
               null;
         end;
      end if;
      Commit (Mutex);
   end Unname;

   procedure Unname
             (  Storage : in out Indexed_Storage_Object;
                Object  : in out Deposit_Handles.Handle
             )  is
      This : Deposit_Ptr := Ptr (Object);
   begin
      if Is_In (Storage.Index, This) then
         declare
            Mutex : Write_Mutex (Storage.Index.Storage);
            ID    : Key := Get (Storage.Index, This);
         begin
            Unname (Storage.Index, This); 
            Unname (Storage.Index.Storage.all, ID);
            Commit (Mutex);
         end;
      end if;
   end Unname;

end Persistent.Data_Bank.Indexed;