--                                                                    --
--  package                         Copyright (c)  Dmitry A. Kazakov  --
--     Persistent.Data_Bank.Indexed                Luebeck            --
--  Interface                                      Autumn, 2004       --
--                                                                    --
--                                Last revision :  12:32 10 Jun 2003  --
--                                                                    --
--  This  library  is  free software; you can redistribute it and/or  --
--  modify it under the terms of the GNU General Public  License  as  --
--  published by the Free Software Foundation; either version  2  of  --
--  the License, or (at your option) any later version. This library  --
--  is distributed in the hope that it will be useful,  but  WITHOUT  --
--  ANY   WARRANTY;   without   even   the   implied   warranty   of  --
--  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU  --
--  General  Public  License  for  more  details.  You  should  have  --
--  received  a  copy  of  the GNU General Public License along with  --
--  this library; if not, write to  the  Free  Software  Foundation,  --
--  Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.    --
--                                                                    --
--  As a special exception, if other files instantiate generics from  --
--  this unit, or you link this unit with other files to produce  an  --
--  executable, this unit does not by  itself  cause  the  resulting  --
--  executable to be covered by the GNU General Public License. This  --
--  exception  does not however invalidate any other reasons why the  --
--  executable file might be covered by the GNU Public License.       --
--____________________________________________________________________--
--
--  This generic package implements persistent storage  interface  using
--  the  operations  defined  in Persistent.Data_Bank. The abstract type
--  Indexed_Storage_Object  can  be used as the base type for a concrete
--  implementation  of  a  data  base  interface. The derived type shall
--  implement the following operations:
--
--  of Persistent.Data_Bank: 
--
--  (o)  Commit a transaction
--  (o)  Delete to delete an object by key
--  (o)  Find to find an object by name
--  (o)  Get_Data to get object's description string
--  (o)  Get_Dependant to enumerate dependants with backward links
--  (o)  Has_Dependants to check if an object can be deleted
--  (o)  Is_In to check if the object in the storage, by key
--  (o)  Rename to set the object name
--  (o)  Roll_Back a transaction
--  (o)  Seize_Read to initiate a read-only transaction
--  (o)  Seize_Write to initiate a read/write transaction
--  (o)  Store to create a new persistent object
--  (o)  Unname to remove object's name
--  (o)  Update to modify a persistent object
--
--  of Persistent:
--
--  (o)  Get_List to get the list of all persistent objects
--
with Ada.Unchecked_Conversion;
with Persistent.Data_Bank.Index;

generic
   type Data_Bank is abstract new Data_Bank_Object with private;
   type Key is new Persistent_Key with private;
package Persistent.Data_Bank.Indexed is
   type Indexed_Storage_Object is abstract new Data_Bank with private;
--
-- Add_To_Index
--
--    Storage  - The indexed storage
--    Object   - To add to the index
--
-- This procedure adds Object to the index. Nothing happens if object is
-- already  in  the  index. This procedure is basically same as Put, but
-- used internally with no interlocking. 
--
-- Exceptions :
--
--    Constraint_Error - Invalid handle
--    Data_Error       - Data base error
--
   procedure Add_To_Index
             (  Storage : in out Indexed_Storage_Object;
                Object  : in out Deposit_Handle
             );
--
-- Create -- A resident object
--
--    Storage        - The indexed storage
--    Object         - The object (a handle to)
--    ID             - The object key (output)
--    Name           - The name the object will have in the storage
--    Class          - The object class
--    Data           - The object data
--    Parameters     - Additional parameters
--    Direct_Links   - The set of objects the catalogued one depends on
--    Backward_Links - The set of backward links
--
-- This  procedure is used to create a persistent object resident in the
-- data  storage.  Such  object  is  just  a  proxy  for the data in the
-- storage. The procedure calls Persistent.Data_Bank.Store and then puts
-- the object in the index, to make Object bound.
--
-- Exceptions :
--
--    Constraint_Error - Invalid handle
--    Data_Error       - Inconsistent data base
--    Name_Error       - Name conflict
--
   procedure Create
             (  Storage        : in out Indexed_Storage_Object;
                Object         : Deposit_Handle;
                ID             : out Key;
                Name           : Wide_String;
                Class          : String;
                Data           : String;
                Parameters     : String;
                Direct_Links   : Deposit_Set;
                Backward_Links : Deposit_Set              
            );
--
-- Get -- Implements Persistent...
--
-- Effects :
--
--    Seize_Read
--
   function Get
            (  Storage : access Indexed_Storage_Object;
               Name    : Wide_String
            )  return Deposit_Handle;
--
-- Get -- Implements Persistent.Data_Bank...
--
   function Get
            (  Storage : access Indexed_Storage_Object;
               Key     : Persistent_Key'Class
            )  return Deposit_Handle;
--
-- Get -- Object's key
--
--    Storage  - The indexed storage
--    Object   - The object which key is requested (a pointer to)
--
-- This  function is used to get a key of the persistent object bound to
-- a  memory-mapped  one.  No  attempt  to  bind  them   is   made,   so
-- Constraint_Error is returned if Object is not bound or null.  
--
-- Returns :
--
--    The object's key in Storage
--
-- Exceptions :
--
--    Constraint_Error - The object is not persistent in Storage
--
   function Get_Key
            (  Storage : Indexed_Storage_Object;
               Object  : Deposit_Ptr
            )  return Key;
--
-- Get_Name -- Implements Persistent...
--
   function Get_Name
            (  Storage : access Indexed_Storage_Object;
               Object  : Deposit_Handle
            )  return Wide_String;
--
-- Is_In -- Implements Persistent...
--
-- Effects :
--
--    Seize_Read
--
   function Is_In
            (  Storage : access Indexed_Storage_Object;
               Name    : Wide_String
            )  return Boolean;
   function Is_In
            (  Storage : access Indexed_Storage_Object;
               Object  : Deposit_Handle
            )  return Boolean;
--
-- Is_Indexed -- Check if a name is in the index
--
--    Storage  - The indexed storage
--    Name     - To check
--
-- Returns :
--
--    True if Name is in the Storage's index
--
   function Is_Indexed
            (  Storage : access Indexed_Storage_Object;
               Name    : Wide_String
            )  return Boolean;
--
-- Purge -- The catalogue of bound objects
--
--    Storage  - The indexed storage
--
-- This procedure is called  upon  destruction  of  a  derived  strorage
-- object.  It  deletes all records of any remaining bound objects. Such
-- objects can be only ones non-resident in the  storage.  The  resident
-- ones should contain a handle to the storage  object  preventing  that
-- from destruction. Purge should be the first procedure of Finalize  to
-- ensure accessibility of the storage for updating. 
--
-- Effects :
--
--    Updating bound objects in the storage
--
   procedure Purge (Storage : in out Indexed_Storage_Object'Class);
--
-- Put -- Implements Persistent...
--
-- Effects :
--
--    Seize_Write
--
   procedure Put
             (  Storage : in out Indexed_Storage_Object;
                Object  : in out Deposit_Handle;
                Name    : Wide_String
             );
   procedure Put
             (  Storage : in out Indexed_Storage_Object;
                Object  : in out Deposit_Handle
             );
--
-- Put -- Implements Persistent.Data_Bank...
--
   procedure Put
             (  Storage : in out Indexed_Storage_Object;
                Key     : Persistent_Key'Class;
                Object  : Deposit'Class
             );
--
-- Rename -- Implements Persistent...
--
-- Effects :
--
--    Seize_Write
--
   procedure Rename
             (  Storage  : in out Indexed_Storage_Object;
                Object   : Wide_String;
                New_Name : Wide_String
             );
   procedure Rename
             (  Storage  : in out Indexed_Storage_Object;
                Object   : in out Deposit_Handle;
                New_Name : Wide_String
             );
--
-- Unname -- Implements Persistent...
--
-- Effects :
--
--    Seize_Write
--
   procedure Unname
             (  Storage : in out Indexed_Storage_Object;
                Object  : Wide_String
             );
   procedure Unname
             (  Storage : in out Indexed_Storage_Object;
                Object  : in out Deposit_Handle
             );

private
   package Storage_Index is
      new Persistent.Data_Bank.Index (Data_Bank, Key);

   type Indexed_Storage_Object is abstract new Data_Bank with record
      Index : Storage_Index.Catalogue
                 (Indexed_Storage_Object'Unchecked_Access);
   end record;

   type Indexed_Storage_Object_Ptr is
      access all Indexed_Storage_Object'Class;
   function To_Persistent_Storage_Ptr is
      new Ada.Unchecked_Conversion
          (  Indexed_Storage_Object_Ptr,
             Storage_Object_Ptr
          );

end Persistent.Data_Bank.Indexed;