--                                                                    --
--  package                         Copyright (c)  Dmitry A. Kazakov  --
--     Persistent.Generic_ODBC                     Luebeck            --
--  Interface                                      Winter, 2002       --
--                                                                    --
--                                Last revision :  12:32 10 Jun 2003  --
--                                                                    --
--  This  library  is  free software; you can redistribute it and/or  --
--  modify it under the terms of the GNU General Public  License  as  --
--  published by the Free Software Foundation; either version  2  of  --
--  the License, or (at your option) any later version. This library  --
--  is distributed in the hope that it will be useful,  but  WITHOUT  --
--  ANY   WARRANTY;   without   even   the   implied   warranty   of  --
--  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU  --
--  General  Public  License  for  more  details.  You  should  have  --
--  received  a  copy  of  the GNU General Public License along with  --
--  this library; if not, write to  the  Free  Software  Foundation,  --
--  Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.    --
--                                                                    --
--  As a special exception, if other files instantiate generics from  --
--  this unit, or you link this unit with other files to produce  an  --
--  executable, this unit does not by  itself  cause  the  resulting  --
--  executable to be covered by the GNU General Public License. This  --
--  exception  does not however invalidate any other reasons why the  --
--  executable file might be covered by the GNU Public License.       --
--____________________________________________________________________--
--
--  This package provides an implementation of persistent storage  based
--  on an ODBC conform data base. The data base sturcture is defined  as
--  follows: 
--
--  The table of objects has the following columns:
--
--  object_id      : a unique object identifier, primary key
--  catalogue_name : a unique object name, Unicode string
--  class_name     : object class, Latin-1 string
--  object_data    : object representation, Latin-1 string
--  parameter_list : object parameters, Latin-1 string
--  created_at     : creation time
--
--  The table of links represent depends on relation:
--
--  dependant : dependent object identifier
--  referent  : reference object identifier
--
--  Each record represents dependency of the  object  identified  by  id
--  from the object identified by reference. Objects that  are  excluded
--  from  garbage  collection  contain  records  referencing the objects
--  themselves.
--
with Ada.Strings.Unbounded;        use Ada.Strings.Unbounded;
with GNU.DB.SQLCLI;                use GNU.DB.SQLCLI;
with GNU.DB.SQLCLI.API;            use GNU.DB.SQLCLI.API;
with GNU.DB.SQLCLI.API.Keys;       use GNU.DB.SQLCLI.API.Keys;
with GNU.DB.SQLCLI.API.Keys.Sets;  use GNU.DB.SQLCLI.API.Keys.Sets;
with Object.Archived.Sets;         use Object.Archived.Sets;
with Persistent.Data_Bank;         use Persistent.Data_Bank;
with Persistent.Handle;            use Persistent.Handle;

with Generic_Unbounded_Ptr_Array;
with Persistent.Data_Bank.Indexed;

pragma Elaborate_All (GNU.DB.SQLCLI.API);

package Persistent.ODBC is
--
-- Create -- An ODBC persistent storage interface object
--
--    Server_Name - To connect to
--    User_Name   - To connect as
--    Password    - Of the user
--    Erase       - Erase the data base upon connecting
--
-- This function creates an ODBC persistent storage interface object and
-- returns  a  handle  to  it. The object is responsible for interacting
-- with  an  ODBC  conform data base. A connection is established to the
-- server  specified  by  the  parameter  Server_Name.  The   parameters
-- User_Name and Password identify the data base user. Refer to the ODBC
-- driver  configuration for further information about these parameters.
-- The parameter Erase when  True  erases  the  data  base  contents  by
-- dropping  all  the tables used for storing persistent objects. If the
-- data base contains any additional tables, they remain untouched.  
--
-- Returns :
--
--    Handle to the created ODBC storage object
--
-- Exceptions :
--
--    Use_Error  - Password or other data might be wrong
--    Data_Error - Data base error
--
   function Create
            (  Server_Name : Wide_String;
               User_Name   : Wide_String;
               Password    : Wide_String;
               Erase       : Boolean := False
            )  return Storage_Handle;
--
-- Disable_Tracing -- Stop tracing
--
--    Storage - A handle to persistent storage
--
-- Exceptions :
--
--    Constraint_Error - Not a valid handle to an ODBC storage
--
   procedure Disable_Tracing
             (  Storage : in out Storage_Handle
             );
--
-- Enable_Tracing -- Set a trace file
--
--    Storage - A handle to persistent storage
--    Name    - Of the trace file
--
-- Exceptions :
--
--    Constraint_Error - Not a valid handle to an ODBC storage
--    Data_Error       - Any other error
--
   procedure Enable_Tracing
             (  Storage : in out Storage_Handle;
                Name    : String
             );
--
-- Get_Connection -- Get a handle to the ODBC connection
--
--    Storage - A handle to persistent storage
--
-- Returns :
--
--    Pointer to the connection object
--    
-- Exceptions :
--
--    Constraint_Error - Not a valid handle to an ODBC storage
--
   function Get_Connection (Storage : Storage_Handle)
      return ODBC_Connection_Ptr;
--
-- Is_ODBC -- Test if the storage handle refers ODBC storage
--
--    Storage - A handle to persistent storage
--
-- Returns :
--
--    True if Storage refers an ODBC one
--
   function Is_ODBC (Storage : Storage_Handle) return Boolean;
   pragma Inline (Is_ODBC);
--
-- Serializable -- Check if concurrent access is safe
--
--    Storage - A handle to persistent storage
--
-- When  the  result  is  false  it  is  unsafe  to  have  more than one
-- connection to the data base. 
--
-- Returns :
--
--    True if the driver supports interlocking
--    
-- Exceptions :
--
--    Constraint_Error - Not a valid handle to an ODBC storage
--
   function Serializable (Storage : Storage_Handle)
      return Boolean;
--
-- Unicode -- Native support
--
--    Storage - A handle to persistent storage 
--
-- Returns :
--
--    True if the driver natively supports Unicode
--    
-- Exceptions :
--
--    Constraint_Error - Not a valid handle to an ODBC storage
--
   function Unicode (Storage : Storage_Handle)
      return Boolean;

private
--
-- The ODBC environment for all connections
--
   Environment : aliased ODBC_Environment;
--
-- Table names:
--
   Direct_Links_Table   : aliased String := "direct_links";
   Backward_Links_Table : aliased String := "backward_links";
   Objects_Table        : aliased String := "objects";
--
-- Columns:
--
   ID_Column           : constant String := "object_id";
   Name_Column         : constant String := "catalogue_name";
   Class_Column        : constant String := "class_name";
   Data_Column         : constant String := "object_data";
   Parameters_Column   : constant String := "parameters_list";
   Time_Column         : constant String := "created_at";
--
-- Prefixes of training set's table names:
--
   Has_In_Table        : constant String := "has_in_";
   Has_Out_Table       : constant String := "has_out_";
   Has_Not_Table       : constant String := "has_not_";
   Has_Not_Out_Table   : constant String := "has_not_out_";
--
-- Arrays of string pointers
--
   type String_Ptr is access String;
   type String_Ptr_Array is array (Integer range <>) of String_Ptr;
   package String_List is
      new Generic_Unbounded_Ptr_Array
          (  Index_Type            => Integer,
             Object_Type           => String,
             Object_Ptr_Type       => String_Ptr,
             Object_Ptr_Array_Type => String_Ptr_Array
          );
   use String_List;

   package Indexed is
      new Persistent.Data_Bank.Indexed (Data_Bank_Object, Object_Key);
   use Indexed; 
--
-- SQL_Type_Info -- Description of the SQL type available for use
--
   type SQL_Type_Info is record
      Data_Type : SQL_Data_Type;
      Name      : Unbounded_String;
   end record;
--
-- Sharing_Type -- The access level to the data base
--
   type Sharing_Type is (Fully, Read_Only, Read_Write);
--
-- Data_Base_Object -- ODBC data base as persistent storage
--
   type Data_Base_Object is new Indexed_Storage_Object with record
      Connection   : ODBC_Connection_Ptr;
      Command      : ODBC_Command_Ptr;
      Shared       : Sharing_Type := Fully;
      Serializable : Boolean      := True;
      Cached_Key   : Object_ID    := No_ID;
      Cached_Set   : Sets.Set;
      Next_ID      : Object_ID;
   --
   -- Driver specifics :
   --
      Unicode          : Boolean;         -- Supports Unicode
      Integer_SQL_Type : SQL_Type_Info;   -- Integer type
      ID_SQL_Type      : SQL_Type_Info;   -- Objects IDs
      Name_SQL_Type    : SQL_Type_Info;   -- Objects names
      Data_SQL_Type    : SQL_Type_Info;   -- Variable strings
      Stamp_SQL_Type   : SQL_Type_Info;   -- Time stamps
   end record;
   type Data_Base_Ptr is access Data_Base_Object'Class;
--
-- Commit -- Overrides Persistent.Data_Bank...
--
   procedure Commit (Storage : in out Data_Base_Object);
--
-- Connect -- Make ODBC persistent storage interface object ready
--
--    Storage     - To prepare to work
--    Server_Name - To connect to
--    User_Name   - To connect as
--    Password    - Of the user
--    Erase       - Erase the data base upon connecting
--
-- This procedure is called from Create to make Storage ready to use. If
-- any type is derived from Data_Base_Object, it should call Connect  at
-- some point after Initialize, but before any use.
--
-- Exceptions :
--
--    Use_Error  - Password or other data might be wrong
--    Data_Error - Data base error
--
   procedure Connect
             (  Storage     : in out Data_Base_Object'Class;
                Server_Name : Wide_String;
                User_Name   : Wide_String;
                Password    : Wide_String;
                Erase       : Boolean := False
             );
--
-- Delete -- Overrides Persistent.Data_Bank...
--
   procedure Delete
             (  Storage : in out Data_Base_Object;
                Key     : Persistent_Key'Class
             );
--
-- Finalize -- Destruction
--
--    Storage - To finalize
--
   procedure Finalize (Storage : in out Data_Base_Object);
--
-- Find -- Overrides Persistent.Data_Bank...
--
   function Find
            (  Storage : access Data_Base_Object;
               Name    : Wide_String
            )  return Persistent_Key'Class;
--
-- Get_Data -- Overrides Persistent.Data_Bank...
--
   procedure Get_Data
             (  Storage    : in out Data_Base_Object;
                Key        : Persistent_Key'Class;
                Class      : out Unbounded_String;
                Data       : out Unbounded_String;
                Parameters : out Unbounded_String
             );
--
-- Get_Dependant -- Overrides Persistent.Data_Bank...
--
   function Get_Dependant
            (  Storage : access Data_Base_Object;
               Key     : Persistent_Key'Class;
               No      : Positive
            )  return Persistent_Key'Class;
--
-- Get_List -- Overrides Persistent...
--
-- Effects :
--
--    Seize_Read
--
   function Get_List
            (  Storage     : access Data_Base_Object;
               Prefix      : Wide_String := "";
               Suffix      : Wide_String := "";
               Equivalence : Wide_Character_Mapping := Identity
            )  return Catalogue.Set;
--
-- Get_Name -- Overrides Persistent.Data_Bank...
--
   function Get_Name
            (  Storage : access Data_Base_Object;
               Key     : Persistent_Key'Class
            )  return Wide_String;
--
-- Has_Dependants -- Overrides Persistent.Data_Bank...
--
   function Has_Dependants
            (  Storage : access Data_Base_Object;
               Key     : Persistent_Key'Class
            )  return Boolean;
--
-- Is_In -- Overrides Persistent.Data_Bank...
--
   function Is_In
            (  Storage : access Data_Base_Object;
               Key     : Persistent_Key'Class
            )  return Boolean;
--
-- Rename -- Overrides Persistent.Data_Bank...
--
   procedure Rename
             (  Storage : in out Data_Base_Object;
                Key     : Persistent_Key'Class;
                Name    : Wide_String
             );
--
-- RollBack -- Overrides Persistent.Data_Bank...
--
   procedure Roll_Back (Storage : in out Data_Base_Object);
--
-- Seize_Read -- Overrides Persistent.Data_Bank...
--
   procedure Seize_Read (Storage : in out Data_Base_Object);
--
-- Seize_Write -- Overrides Persistent.Data_Bank...
--
   procedure Seize_Write (Storage : in out Data_Base_Object);
--
-- Store -- Overrides Persistent.Data_Bank...
--
   function Store
            (  Storage        : access Data_Base_Object;
               Name           : Wide_String;
               Class          : String;
               Data           : String;
               Parameters     : String;
               Direct_Links   : Deposit_Set;
               Backward_Links : Deposit_Set
            )  return Persistent_Key'Class;
   function Store
            (  Storage        : access Data_Base_Object;
               Class          : String;
               Data           : String;
               Parameters     : String;
               Direct_Links   : Deposit_Set;
               Backward_Links : Deposit_Set
            )  return Persistent_Key'Class;
--
-- Unicode -- Overrides Persistent...
--
--    Storage - A pesistent storage object
--
-- Returns :
--
--    True if Unicode is supported
--
   function Unicode (Storage : Data_Base_Object) return Boolean;
--
-- Unname -- Overrides Persistent.Data_Bank...
--
   procedure Unname
             (  Storage : in out Data_Base_Object;
                Key     : Persistent_Key'Class
             );
--
-- Update -- Overrides Persistent.Data_Bank...
--
   procedure Update
             (  Storage        : in out Data_Base_Object;
                Key            : Persistent_Key'Class;
                Class          : String;
                Data           : String;
                Parameters     : String;
                Direct_Links   : Deposit_Set;
                Backward_Links : Deposit_Set
             );
--
-- Value -- Overrides Persistent.Data_Bank...
--
   function Value
            (  Storage : Data_Base_Object;
               Key     : String
            )  return Persistent_Key'Class;

end Persistent.ODBC;
